local soundstuff = require("soundstuff")

Rocket = {}

function Rocket:new(id, x, y, fuel)
    local rocket = {
        id = id,
        turnspeed = 4000,
        speed = 4900,
        reversespeed = 3000,
        maxFuel = fuel or 120,
        fuelBurnSpeed = 20,
        alive = true
    }
    setmetatable(rocket, self)
    self.__index = self

    rocket.fuel = rocket.maxFuel
    rocket.physicsObj = verlet.newTriangle(x - 20, y + 20, x, y - 30, x + 20,
                                           y + 20)
    rocket.state = Rocket.states.failling

    rocket.engineSparks = love.graphics.newParticleSystem(
                              All_Images.smokeParticle, 100)
    rocket.engineSparks:setParticleLifetime(0.5, 1)

    rocket.engineSparks:setSizes(0.5, 1.5)
    -- rocket.engineSparks:setSpread(0)
    -- rocket.engineSparks:setSpeed(50,80)
    rocket.engineSparks:setColors(1, 1, 1, 0.5, 1, 1, 1, 0) -- Fade to transparency.
    rocket.engineSparks:setLinearAcceleration(0, 0, 0, 0)
    -- rocket.engineSparks:setRotation(-2 * math.pi, 2 * math.pi)
    rocket.offsetX =
        (rocket.physicsObj:center() - rocket.physicsObj.verts[1]).x + 4
    rocket.offsetY =
        (rocket.physicsObj:center() - rocket.physicsObj.verts[2]).y - 1

    rocket.physicsObj.collissionObserver = rocket
    return rocket
end

function Rocket:applyUpgrades(upgradetbl)
    self.maxFuel = self.maxFuel + (15 * upgradetbl.fuel)
    self.fuel = self.maxFuel
    self.turnspeed = self.turnspeed + (200 * upgradetbl.thrusters)
    self.reversespeed = self.reversespeed * (upgradetbl.reverse / 3)
end

function Rocket:render()
    love.graphics.setColor(1, 1, 1, 1)
    local verts = self.physicsObj.verts
    love.graphics.draw(self.engineSparks, 0, 0)
    local t = verts[2]
    local c = self.physicsObj:center()
    local d = t - c
    local angle = math.atan2(d.y, d.x) + math.pi / 2

    love.graphics.draw(All_Images.ship, self.physicsObj:center().x,
                       self.physicsObj:center().y, angle, 1, 1, self.offsetX,
                       self.offsetY)

end

function Rocket:facing()
    return (self.physicsObj.verts[1] - self.physicsObj.verts[3]):normal()
               :normalize()
end

function Rocket:speedup(dt, amount)
    if self.fuel > 0 then
        verlet.accelerate(self.physicsObj,
                          self:facing():mult(self.speed * amount))
        self.fuel = self.fuel - self.fuelBurnSpeed * amount * dt
        self.engineSparks:setEmissionRate(math.ceil(100 * amount))
        soundstuff.sources.rocket:play()
    end

end

function Rocket:turnright(dt, amount)
    if self.fuel > 0 then
        local steering = self:facing():rotate(math.pi / 2)
        verlet.accelerateSingleVert(self.physicsObj, 2, steering:clone():mult(
                                        self.turnspeed * amount))
        verlet.accelerateSingleVert(self.physicsObj, 3, steering:clone():mult(
                                        -self.turnspeed * amount))
        self.fuel = self.fuel - self.fuelBurnSpeed * amount * (dt / 2)
        self.engineSparks:setEmissionRate(math.ceil(100 * amount))
        soundstuff.sources.rocket:play()
    end

end

function Rocket:turnleft(dt, amount)
    if self.fuel > 0 then
        local steering = self:facing():rotate(math.pi / 2)
        verlet.accelerateSingleVert(self.physicsObj, 2, steering:clone():mult(
                                        -self.turnspeed * amount))
        verlet.accelerateSingleVert(self.physicsObj, 1, steering:clone():mult(
                                        self.turnspeed * amount))
        self.fuel = self.fuel - self.fuelBurnSpeed * amount * (dt / 2)
        self.engineSparks:setEmissionRate(math.ceil(100 * amount))
        soundstuff.sources.rocket:play()
    end

end

function Rocket:reverse(dt, amount)
    if self.fuel > 0 and self.reversespeed > 0 then
        verlet.accelerate(self.physicsObj,
                          self:facing():mult(-self.reversespeed * amount))
        self.fuel = self.fuel - self.fuelBurnSpeed * amount * dt
        self.engineSparks:setEmissionRate(math.ceil(100 * amount))
        soundstuff.sources.rocket:play()
    end
end

function Rocket:update(dt)

    local verts = self.physicsObj.verts
    local emitterDir = (verts[1] - verts[3]):normal():mult(-1)
    self.engineSparks:setLinearAcceleration(emitterDir.x - 200,
                                            emitterDir.y * 10,
                                            emitterDir.x + 200,
                                            emitterDir.y * 10)
    local sparkEmitingPos = {
        x = (verts[1].x + verts[3].x) / 2,
        y = (verts[1].y + verts[3].y) / 2
    }
    self.engineSparks:setPosition(sparkEmitingPos.x, sparkEmitingPos.y)
    self.engineSparks:update(dt)

    local center = self.physicsObj:center()
    if camera then camera:setPosition(center.x, center.y) end

    self.state.update(self, dt)

end

function Rocket:notmoving()
    self.engineSparks:setEmissionRate(0)
    soundstuff.sources.rocket:pause()
end

function Rocket:refuel(amount)
    self.fuel = math.min(self.fuel + amount, self.maxFuel)
end

function Rocket:oncolission(info)

    if self.alive then

        game.removeObj(self)
        self.alive = false
        gameover(false, "You crashed")
        soundstuff.sources.rocketExplosion:play()
        self:notmoving()

        player = nil
        game.createExplosion(self.physicsObj:center().x,
                             self.physicsObj:center().y)
        self:createWreck()

    end
end

Rocket.states = {}
Rocket.states.failling = {}
Rocket.states.failling.update = function(obj, dt)
    local currentVelocity1 = obj.physicsObj.verts[1] - obj.physicsObj.last[1]
    local currentVelocity2 = obj.physicsObj.verts[2] - obj.physicsObj.last[2]
    local currentVelocity3 = obj.physicsObj.verts[3] - obj.physicsObj.last[3]

    local targetVelocity = (currentVelocity1 + currentVelocity2 +
                               currentVelocity3):mult(1 / 3)

    local steeringVert1 = targetVelocity - currentVelocity1
    local steeringVert2 = targetVelocity - currentVelocity2
    local steeringVert3 = targetVelocity - currentVelocity3

    verlet.accelerateSingleVert(obj.physicsObj, 1, steeringVert1:mult(10000))
    verlet.accelerateSingleVert(obj.physicsObj, 2, steeringVert2:mult(10000))
    verlet.accelerateSingleVert(obj.physicsObj, 3, steeringVert3:mult(10000))
end

function Rocket.createWreck(rocket)
    local o1 = {id = game._cnt}
    o1.physicsObj = verlet.newTriangle(rocket.physicsObj.verts[1].x,
                                       rocket.physicsObj.verts[1].y,
                                       rocket.physicsObj.verts[2].x,
                                       rocket.physicsObj.verts[2].y,
                                       (rocket.physicsObj.verts[1].x +
                                           rocket.physicsObj.verts[3].x) / 2,
                                       (rocket.physicsObj.verts[1].y +
                                           rocket.physicsObj.verts[3].y) / 2)
    o1.physicsObj.last = {
        rocket.physicsObj.last[1]:clone(), rocket.physicsObj.last[2]:clone(),
        Vector:new(
            (rocket.physicsObj.last[1].x + rocket.physicsObj.last[3].x) / 2,
            (rocket.physicsObj.last[1].y + rocket.physicsObj.last[3].y) / 2)
    }
    o1.offsetX = 12.5 + 3
    o1.offsetY = 25 + 5

    local wreckfacing1 = function(self)
        return (self.physicsObj.verts[2] - self.physicsObj.verts[3]):normalize()
    end

    o1.image = All_Images.shipwreck1
    o1.wreckfacing = wreckfacing1
    local renderfunc1 = function(self)

        love.graphics.setColor(1, 1, 1, 1)
        local verts = self.physicsObj.verts
        local d = self:wreckfacing()
        local angle = math.atan2(d.y, d.x) + math.pi / 2
        love.graphics.draw(self.image, self.physicsObj:center().x,
                           self.physicsObj:center().y, angle, 1, 1,
                           self.offsetX, self.offsetY)

    end
    o1.render = renderfunc1

    table.insert(game._objectsToSpawn, o1)

    local o2 = {id = game._cnt}

    local renderfunc2 = function(self)
        love.graphics.setColor(1, 1, 1, 1)
        local verts = self.physicsObj.verts
        local d = self:wreckfacing()
        local angle = math.atan2(d.y, d.x) + math.pi / 2
        -- TODO: center  koordinate stimm vorne und hinten nicht aber drunter passt sie wtf=
        love.graphics.draw(self.image, self.physicsObj:center().x,
                           self.physicsObj:center().y, angle, 1, 1,
                           self.offsetX, self.offsetY)

    end
    o2.wreckfacing = function(self)
        return (self.physicsObj.verts[2] - self.physicsObj.verts[1]):normalize()
    end

    o2.physicsObj = verlet.newTriangle((rocket.physicsObj.verts[1].x +
                                           rocket.physicsObj.verts[3].x) / 2,
                                       (rocket.physicsObj.verts[1].y +
                                           rocket.physicsObj.verts[3].y) / 2,
                                       rocket.physicsObj.verts[2].x,
                                       rocket.physicsObj.verts[2].y,
                                       rocket.physicsObj.verts[3].x,
                                       rocket.physicsObj.verts[3].y)
    o2.physicsObj.last = {
        Vector:new(
            (rocket.physicsObj.verts[1].x + rocket.physicsObj.verts[3].x) / 2,
            (rocket.physicsObj.verts[1].y + rocket.physicsObj.verts[3].y) / 2),
        rocket.physicsObj.last[2]:clone(), rocket.physicsObj.last[3]:clone()
    }

    o2.offsetX = 12.5 - 3
    o2.offsetY = 25 + 5
    o2.image = All_Images.shipwreck2

    o2.render = renderfunc2

    table.insert(game._objectsToSpawn, o2)
end
