require("vector")
require("sat")
require("utils")
verlet = {}

verlet._cnt = 1
verlet.objects = {}
verlet._objectsToSpawn = {}
verlet._objectsToDie = {}
verlet._movingObjects = {}
verlet.gravity = Vector:new(0, 100)

verlet._center = Vector:new(500, 500)
verlet._radius = 450
verlet.substeps = 8
verlet.response_coef = 0.75
verlet.links = {}
verlet.magnetStrength = 500
verlet.magnet = nil

function verlet.reset()
    verlet._cnt = 1
    verlet.objects = {}
    verlet._objectsToSpawn = {}
    verlet._objectsToDie = {}
end

function verlet.removeObj(obj) table.insert(verlet._objectsToDie, obj) end

function verlet._cachedCenter(object) return object._cachedCenter end

function verlet.newCircle(x, y, radius)
    local createdObj = {
        id = verlet._cnt,
        pos = Vector:new(x, y),
        radius = radius,
        accel = Vector:new(0, 0),
        color = {r = 1, g = 2, b = 1, a = 1},
        type = "c"
    }
    table.insert(verlet._objectsToSpawn, createdObj)
    verlet._cnt = verlet._cnt + 1
    createdObj.last = createdObj.pos:clone()
    createdObj.center = function(self) return self.pos:clone() end

    createdObj.mass = function(self)
        return self.radius * self.radius * math.pi
    end

    createdObj.colissioncheckradius = createdObj.radius

    return createdObj

end

function verlet.newTriangle(x1, y1, x2, y2, x3, y3)
    local createdObj = {
        id = verlet._cnt,
        color = {r = 1, g = 2, b = 1, a = 1},
        type = "t"
    }
    table.insert(verlet._objectsToSpawn, createdObj)
    verlet._cnt = verlet._cnt + 1
    createdObj.verts = {
        Vector:new(x1, y1), Vector:new(x2, y2), Vector:new(x3, y3)
    }

    createdObj.calcCenter = function(self)
        return Vector:new((createdObj.verts[1].x + createdObj.verts[2].x +
                              createdObj.verts[3].x) / 3,
                          (createdObj.verts[1].y + createdObj.verts[2].y +
                              createdObj.verts[3].y) / 3)
    end
    createdObj.center = verlet._cachedCenter
    createdObj._cachedCenter = createdObj:calcCenter()

    createdObj.lengths = {
        (createdObj.verts[1] - createdObj.verts[2]):length(),
        (createdObj.verts[2] - createdObj.verts[3]):length(),
        (createdObj.verts[3] - createdObj.verts[1]):length()
    }

    createdObj.last = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.last, createdObj.verts[i]:clone())
    end
    createdObj.accel = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.accel, Vector:new(0, 0))
    end

    createdObj.mass = function(self)
        local a = (self.verts[2] - self.verts[1]):length()
        local b = (self.verts[3] - self.verts[2]):length()
        local c = (self.verts[1] - self.verts[3]):length()
        local s = (a + b + c) / 2

        return math.sqrt(s * (s - a) * (s - b) * (s - c))
    end

    createdObj.colissioncheckradius = createdObj.verts[1]:distance(
                                          createdObj.verts[2])

    return createdObj
end

function verlet.newOctagon(centerX, centerY, radius)
    local createdObj = {
        id = verlet._cnt,
        color = {r = 1, g = 2, b = 1, a = 1},
        type = "o"
    }
    table.insert(verlet._objectsToSpawn, createdObj)
    verlet._cnt = verlet._cnt + 1
    createdObj.verts = {}
    createdObj.radius = radius
    local normalVec = Vector:new(1, 0)
    local centerpos = Vector:new(centerX, centerY)
    for r = 0, math.pi * 2 - math.pi / 4, math.pi / 4 do
        local currentVert = normalVec:clone():rotate(r):mult(radius):add(
                                centerpos)
        table.insert(createdObj.verts, currentVert)
    end
    local xsum = 0
    local ysum = 0
    for i, v in ipairs(self.verts) do
        xsum = xsum + v.x
        ysum = ysum + v.y
    end
    createdObj.calcCenter = function(self)
        return Vector:new(xsum / #self.verts, ysum / #self.verts)

    end

    createdObj.center = verlet._cachedCenter

    createdObj._cachedCenter = createdObj:calcCenter()

    createdObj.lengths = {}
    for i = 1, #createdObj.verts do
        createdObj.lengths[i] = {}
        for j = i, #createdObj.verts do
            if j ~= i then
                createdObj.lengths[i][j] =
                    (createdObj.verts[i] - createdObj.verts[j]):length()
            end
        end
    end

    createdObj.last = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.last, createdObj.verts[i]:clone())
    end
    createdObj.accel = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.accel, Vector:new(0, 0))
    end

    createdObj.mass = function(self)
        local a = (self.verts[2] - self.verts[1]):length()
        local b = (self.verts[3] - self.verts[2]):length()
        local c = (self.verts[1] - self.verts[3]):length()
        local s = (a + b + c) / 2

        return math.sqrt(s * (s - a) * (s - b) * (s - c))
    end

    createdObj.colissioncheckradius = createdObj.radius

    return createdObj
end

function verlet.newRect(x, y, width, height, rotation)
    local createdObj = {
        id = verlet._cnt,
        pos = Vector:new(x, y),
        height = height,
        width = width,
        color = {r = 1, g = 2, b = 1, a = 1},
        type = "r",
        rotation = rotation
    }

    table.insert(verlet._objectsToSpawn, createdObj)
    verlet._cnt = verlet._cnt + 1
    createdObj.verts = sat.verts0(createdObj.pos.x, createdObj.pos.y,
                                  createdObj.width, createdObj.height)
    createdObj.calcCenter = function(self)
        return self.verts[1] + (self.verts[3] - self.verts[1]):mult(0.5)
    end
    createdObj.center = verlet._cachedCenter

    createdObj._cachedCenter = createdObj:calcCenter()
    for i, v in ipairs(createdObj.verts) do
        v:rotateAround(rotation, createdObj:center())
    end

    createdObj.last = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.last, createdObj.verts[i]:clone())
    end
    createdObj.accel = {}
    for i = 1, #createdObj.verts do

        table.insert(createdObj.accel, Vector:new(0, 0))
    end

    createdObj.mass = function(self) return self.height * self.width end

    createdObj.colissioncheckradius = math.sqrt(
                                          createdObj.width * createdObj.width +
                                              createdObj.height *
                                              createdObj.height)

    return createdObj

end

function verlet.setCollissionObserver(obj, fun) obj.collissionObserver = fun end

function verlet.link(obj1, obj2)
    table.insert(verlet.links, {
        o1 = obj1,
        o2 = obj2,
        targetDist = (obj1.radius + obj2.radius) + 2
    })
end

function verlet._addAtEndOfMovingObjects(objToSpawn, objs)
    for i = 1, #objs do
        if objs[i].pinned then
            table.insert(objs, i, objToSpawn)
            return
        end
    end
    table.insert(objs, objToSpawn)
end

function verlet._checkObjectsToSpawn()
    for i, v in ipairs(verlet._objectsToSpawn) do
        verlet._addAtEndOfMovingObjects(v, verlet.objects)
    end

    verlet._objectsToSpawn = {}
end

function verlet._remove(obj)
    for i = 1, #verlet.objects do
        if verlet.objects[i].id == obj.id then
            table.remove(verlet.objects, i)
            return
        end
    end

end

function verlet._checkObjectsToDie()
    for i, v in ipairs(verlet._objectsToDie) do verlet._remove(v) end
    verlet._objectsToDie = {}
end

function verlet.game_update(dt)
    verlet._checkObjectsToSpawn()
    verlet._checkObjectsToDie()

    local step_dt = dt / verlet.substeps
    for i = 1, verlet.substeps do

        verlet.apply_gravity(step_dt)
        verlet.apply_effects(step_dt)
        verlet.update_objects(step_dt)
        verlet.solve_collisions(step_dt)
        verlet.apply_constraints(step_dt)

    end

end

function verlet.update_objects(dt)
    for i = 1, #verlet.objects do
        local v = verlet.objects[i]
        if v.pinned then return end
        verlet.apply_friction(v, dt)
        verlet.update(v, dt)
    end
end

function verlet.moveVert(obj, other, vert, amount)
    local to = other:center() - obj:center()
    if to:scalarProduct(amount) > 0 then
        vert:minus(amount)
    else
        vert:add(amount)
    end

end

function verlet.moveRect(obj, ammount)
    for i = 1, #obj.verts do obj.verts[i]:add(ammount) end
end

function verlet.moveEdge(collissionInfo, amount)
    local e1 = collissionInfo.edge.start
    local e2 = collissionInfo.edge.endp
    local collissionVector = amount

    local e1dist = collissionInfo.vert:distance(e1)
    local e2dist = collissionInfo.vert:distance(e2)
    local t = e2dist / (e1dist + e2dist)
    -- local t = nil
    -- if math.abs(e1.x - e2.x) > math.abs(e1.y - e2.y) then
    --     t = (collissionInfo.vert.x - collissionVector.x - e1.x) / (e2.x - e1.x);
    -- else
    --     t = (collissionInfo.vert.y - collissionVector.y - e1.y) / (e2.y - e1.y);
    -- end

    local lambda = 1.0 / (t * t + (1 - t) * (1 - t))
    e1:minus(amount * (t) * lambda)
    e2:minus(amount * (1 - t) * lambda)
end

function verlet.ispoly(obj)
    return obj.type == "r" or obj.type == "t" or obj.type == "o"
end

function verlet.notifyColission(o1, o2, colissioninfo)
    if o1.collissionObserver then
        o1.collissionObserver:oncolission(colissioninfo)
    end
    if o2.collissionObserver then
        o2.collissionObserver:oncolission(colissioninfo)
    end
end

function verlet.satNecessary(o1, o2)
    -- TOOD: only use a center coordinate and no function. Should save us some cpu time as well
    return (not o1.pinned or not o2.pinned) and
               o1:center():distance(o2:center()) <= o1.colissioncheckradius +
               o2.colissioncheckradius

end

function verlet.solve_collision(dt, obj1, obj2)
    if not verlet.satNecessary(obj1, obj2) then
        -- fast track don't do expensive sat check
        return
    end

    if (obj1.type == "c" and obj2.type == "c") then
        local distance = obj1.pos - obj2.pos
        local min_dist = obj1.radius + obj2.radius
        if distance:length() < min_dist then
            local massRatio1 = obj1.radius / (obj1.radius + obj2.radius)
            local massRatio2 = obj2.radius / (obj1.radius + obj2.radius)
            local delta = 0.5 * verlet.response_coef *
                              (distance:length() - min_dist)
            local collissionDir = distance:normalize()
            obj1.pos:minus(collissionDir:clone():mult(massRatio2 * delta))
            obj2.pos:add(collissionDir:clone():mult(massRatio1 * delta))
            verlet.accelerate(obj1,
                              (obj1.last - obj1.pos):mult(0.005 / (dt * dt)))
            verlet.accelerate(obj2,
                              (obj2.last - obj2.pos):mult(0.005 / (dt * dt)))
            verlet.notifyColission(obj1, obj2, colissioninfo)
        end
    elseif (verlet.ispoly(obj1) and verlet.ispoly(obj2)) then

        if obj1.pinned then
            -- I have no idea why but whihout switching the objects in this case the 
            -- collissions sometimes fail spectacularly 
            local tmp = obj1
            obj1 = obj2
            obj2 = tmp
        end

        local colissioninfo = sat.checkoverlap(obj1, obj2)

        if colissioninfo ~= nil then
            local area1 = obj1:mass()
            local area2 = obj2:mass()

            local areatotal = (area1 + area2)
            local delta = 0.5 * verlet.response_coef * colissioninfo.overlap
            if (obj1.pinned or obj2.pinned) and
                (not obj1.pinned or not obj2.pinned) then
                delta = 1 * verlet.response_coef * colissioninfo.overlap
            end

            local to = colissioninfo.edge.obj:center() -
                           colissioninfo.vert.obj:center()
            local scalar = to:scalarProduct(colissioninfo.axis)
            if (scalar > 0) then
                colissioninfo.axis = colissioninfo.axis:mult(-1)
            end

            verlet.moveVert(colissioninfo.vert.obj, colissioninfo.edge.obj,
                            colissioninfo.vert,
                            colissioninfo.axis:clone():mult(
                                delta *
                                    (colissioninfo.edge.obj:mass() / areatotal)))
            verlet.moveEdge(colissioninfo,
                            colissioninfo.axis:clone():mult(
                                delta *
                                    (colissioninfo.vert.obj:mass() / areatotal)))

            for i = 1, #obj1.accel do
                if obj1.verts[i] == colissioninfo.vert.obj then
                    obj1.accel[i]:add((obj1.last[i] - obj1.verts[i]))
                end

            end

            for i = 1, #obj2.accel do
                if obj2.verts[i] == colissioninfo.edge.endp or obj2.verts[i] ==
                    colissioninfo.edge.start then
                    obj2.accel[i]:add((obj2.last[i] - obj2.verts[i]))
                end

            end

            verlet.notifyColission(obj1, obj2, colissioninfo)
        end
    else
        if not verlet.satNecessary(obj1, obj2) then
            -- fast track don't do expensive sat check
            return
        end
        local colissioninfo = nil
        local circle = nil
        local rect = nil
        if verlet.ispoly(obj1) and obj2.type == "c" then
            circle = obj2
            rect = obj1
        elseif obj1.type == "c" and verlet.ispoly(obj2) then
            circle = obj1
            rect = obj2
        else
            -- Fehler
            return
        end

        local colissioninfo = sat.checkoverlapc(rect, circle)

        if colissioninfo ~= nil then

            local circlemass = circle:mass()
            local rectmass = rect:mass()
            local areatotal = (circlemass + rectmass)

            local delta = 0.5 * verlet.response_coef * colissioninfo.overlap

            if (colissioninfo.vert.obj == rect) then
                verlet.moveVert(colissioninfo.vert.obj, colissioninfo.vert,
                                colissioninfo.axis:clone():mult(
                                    delta * (circlemass / areatotal)))
                circle.pos:minus(colissioninfo.axis:clone():mult(
                                     delta * (rectmass / areatotal)))
            else
                verlet.moveEdge(colissioninfo, colissioninfo.axis:clone():mult(
                                    delta * (circlemass / areatotal)))
                circle.pos:add(colissioninfo.axis:clone():mult(
                                   delta * (rectmass / areatotal)))
            end
            verlet.accelerate(circle, (circle.last - circle.pos))

            for i = 1, #rect.accel do
                rect.accel[i]:add((rect.last[i] - rect.verts[i]))
            end

            verlet.notifyColission(obj1, obj2, colissioninfo)

        end

    end
end

function verlet.solve_collisions(dt)

    for i = 1, #verlet.objects do
        local obj1 = verlet.objects[i]
        if obj1.pinned then
            -- pinned  (imobile objects are always at the end of the table. We dont need to check those
            return
        end
        for j = i + 1, #verlet.objects do
            local obj2 = verlet.objects[j]
            verlet.solve_collision(dt, obj1, obj2)
        end
    end
end

function verlet._fixEdge(verts, i1, i2, targetDist)
    local distance = verts[i2] - verts[i1]
    local diff = distance:length() - targetDist

    distance:normalize()

    verts[i1]:add(distance * (diff * 0.5))
    verts[i2]:minus(distance * (diff * 0.5))
end

function verlet.triangleConstraints(dt, obj)
    verlet._fixEdge(obj.verts, 2, 1, obj.lengths[1])
    verlet._fixEdge(obj.verts, 3, 2, obj.lengths[2])
    verlet._fixEdge(obj.verts, 1, 3, obj.lengths[3])

    verlet._fixEdge(obj.last, 2, 1, obj.lengths[1])
    verlet._fixEdge(obj.last, 3, 2, obj.lengths[2])
    verlet._fixEdge(obj.last, 1, 3, obj.lengths[3])
end

function verlet.octagonConstraints(dt, obj)
    for i = 1, #obj.verts do
        for j = i, #obj.verts do
            if i ~= j then
                verlet._fixEdge(obj.verts, i, j, obj.lengths[i][j])
                verlet._fixEdge(obj.last, i, j, obj.lengths[i][j])
            end

        end
    end

end

function verlet.rectConstraints(dt, obj)
    local diag = math.sqrt(obj.width * obj.width + obj.height * obj.height)
    verlet._fixEdge(obj.verts, 1, 2, obj.width)
    verlet._fixEdge(obj.verts, 1, 3, diag)
    verlet._fixEdge(obj.verts, 2, 3, obj.height)
    verlet._fixEdge(obj.verts, 3, 4, obj.width)
    verlet._fixEdge(obj.verts, 4, 2, diag)
    verlet._fixEdge(obj.verts, 4, 1, obj.height)

    verlet._fixEdge(obj.last, 1, 2, obj.width)
    verlet._fixEdge(obj.last, 1, 3, diag)
    verlet._fixEdge(obj.last, 2, 3, obj.height)
    verlet._fixEdge(obj.last, 3, 4, obj.width)
    verlet._fixEdge(obj.last, 4, 2, diag)
    verlet._fixEdge(obj.last, 4, 1, obj.height)

end

function verlet.stayInWorldConstraint(dt, obj)
    if obj.type == "c" then
        local derivationx = verlet.width / 2 - obj.pos.x
        local derivationy = verlet.height / 2 - obj.pos.y
        local wasmoved = false
        if math.abs(derivationx) > (verlet.width / 2 - obj.radius) then
            derivationx = derivationx / math.abs(derivationx)
            obj.pos.x = verlet.width / 2 -
                            (derivationx * (verlet.width / 2 - obj.radius))
            wasmoved = true
        end

        if math.abs(derivationy) > (verlet.height / 2 - obj.radius) then
            derivationy = derivationy / math.abs(derivationy)
            obj.pos.y = verlet.height / 2 -
                            (derivationy * (verlet.height / 2 - obj.radius))
            wasmoved = true
        end

        if wasmoved then
            verlet.accelerate(obj, (obj.last - obj.pos))
            if obj.collissionObserver then
                obj.collissionObserver:oncolission("OutOfBounds")
            end
        end
    elseif verlet.ispoly(obj) then
        -- first, find furtherst corner
        for i = 1, #obj.verts do
            local derivationx = verlet.width / 2 - obj.verts[i].x
            local derivationy = verlet.height / 2 - obj.verts[i].y
            local wasmoved = false
            if math.abs(derivationx) > (verlet.width / 2) then
                derivationx = derivationx / math.abs(derivationx)
                obj.verts[i].x = verlet.width / 2 -
                                     (derivationx * (verlet.width / 2))
                wasmoved = true
            end

            if math.abs(derivationy) > (verlet.height / 2) then
                derivationy = derivationy / math.abs(derivationy)
                obj.verts[i].y = verlet.height / 2 -
                                     (derivationy * (verlet.height / 2))
                wasmoved = true
            end

            if wasmoved then
                for i = 1, #obj.accel do
                    obj.accel[i]:add(obj.last[i] - obj.verts[i])
                end
                if obj.collissionObserver then
                    obj.collissionObserver:oncolission("OutOfBounds")
                end
            end

        end

    end
end

function verlet.apply_constraints(dt)
    for i = 1, #verlet.objects do
        local v = verlet.objects[i]
        if v.pinned then
            return
        end
        verlet.stayInWorldConstraint(dt, v)
        -- verlet.center_constraint(dt, v)
        verlet.pinConstraints(dt, v)

        if v.type == "r" then
            verlet.rectConstraints(dt, v)
        elseif v.type == "t" then
            verlet.triangleConstraints(dt, v)
        elseif v.type == "o" then
            verlet.octagonConstraints(dt, v)
        end

    end
    for i, v in ipairs(verlet.links) do verlet.chainConstraints(dt, v) end

end

function verlet._verts(obj)
    return sat.verts(obj.pos.x, obj.pos.y, obj.width, obj.height, obj.rotation,
                     obj:center())
end

function verlet.center_constraint(dt, obj)
    if obj.type == "c" then
        local distance_to_center = verlet._center - obj.pos
        local len = distance_to_center:length()
        if len > (verlet._radius - obj.radius) then
            local dis_to_move = distance_to_center:normalize()
            obj.pos = verlet._center -
                          (dis_to_move:mult(verlet._radius - obj.radius))
        end
    elseif verlet.ispoly(obj) then
        -- first, find furtherst corner
        for i = 1, #obj.verts do
            local distance_to_center = verlet._center - obj.verts[i]
            local len = distance_to_center:length()
            if len > verlet._radius then
                local dis_to_move = distance_to_center:normalize()
                obj.verts[i] = verlet._center -
                                   (dis_to_move:mult(verlet._radius))
            end
        end

    end

end

function verlet.pinConstraints(dt, obj)
    if obj.pinned == true then
        if obj.type == "c" and obj.pinnedTo:distance(obj.pos) > 0 then
            obj.pos = obj.pinnedTo:clone()
        elseif verlet.ispoly(obj) then
            for i = 1, #obj.verts do
                obj.verts[i] = obj.pinnedTo[i]:clone()
            end
        end

    end
end

function verlet.pin(obj, pos)
    obj.pinned = true
    if obj.type == "c" then
        obj.pinnedTo = pos:clone()
    elseif verlet.ispoly(obj) then
        obj.pinnedTo = {}
        for i = 1, #obj.verts do
            table.insert(obj.pinnedTo, obj.verts[i]:clone())
        end
    end
    obj._cachedCenter = obj:center()
    obj.center = verlet._cachedCenter

end

function verlet.unpin(obj)
    obj.pinned = false
    obj.pinnedTo = nil
end

function verlet.chainConstraints(dt, link)
    local axis = link.o1.pos - link.o2.pos
    local dist = axis:length()
    local n = axis:normalize()
    local delta = link.targetDist - dist

     if delta > 0 then return end
    local massRatio1 = link.o1.radius / (link.o1.radius + link.o2.radius)
    local massRatio2 = link.o2.radius / (link.o1.radius + link.o2.radius)
    link.o1.pos:add(n:clone():mult(delta * massRatio2))
    link.o2.pos:minus(n:clone():mult(delta * massRatio1))
end

function verlet.update(obj, dt)
    if obj.type == "c" then
        local displacement = obj.pos - obj.last
        obj.last:set(obj.pos)
        obj.pos:add(displacement + (obj.accel:mult(dt * dt)))
        obj.accel:setxy(0, 0)
    else
        if verlet.ispoly(obj) then
            for i = 1, #obj.verts do
                local displacement = obj.verts[i] - obj.last[i]
                obj.last[i]:set(obj.verts[i])
                obj.verts[i]:add(displacement + (obj.accel[i]:mult(dt * dt)))
                obj.accel[i]:setxy(0, 0)
                obj._cachedCenter = obj:calcCenter()
            end

        end
    end

end

function verlet.apply_friction(obj, dt)

    if obj.type == "c" then
        local vec = obj.last - obj.pos
        verlet.accelerate(obj, vec:mult(0.3))
    elseif verlet.ispoly(obj) then
        for i, v in ipairs(obj.verts) do
            local vec = obj.last[i] - v
            verlet.accelerateSingleVert(obj, i, vec:mult(0.5))
        end

    end
end

function verlet.accelerate(obj, vec)
    if obj.type == "c" then
        obj.accel:add(vec)
    elseif verlet.ispoly(obj) then
        for i = 1, #obj.accel do obj.accel[i]:add(vec) end

    end

end

function verlet.accelerateSingleVert(obj, vertIdx, vec)
    if obj.type == "c" then
        obj.accel:add(vec)
    elseif verlet.ispoly(obj) then
        obj.accel[vertIdx]:add(vec)

    end

end

function verlet.apply_gravity()
    for i = 1, #verlet.objects do
        local v = verlet.objects[i]
        if v.pinned then
            return
        end
        verlet.accelerate(v, verlet.gravity)
    end
end

function verlet.apply_effects()
    if verlet.magnet ~= nil then
        for i, v in ipairs(verlet.objects) do
            local distance = verlet.magnet - v:center()
            verlet.accelerate(v, distance:normalize() * verlet.magnetStrength)
        end
    end

end

function verlet.render()

    for i, v in ipairs(verlet.objects) do
        love.graphics.setColor(v.color.r, v.color.g, v.color.b, v.color.a)
        if v.type == "c" then
            love.graphics.circle("fill", v.pos.x, v.pos.y, v.radius)
        elseif v.type == "r" then
            local verts = v.verts
            love.graphics.polygon("fill", verts[1].x, verts[1].y, verts[2].x,
                                  verts[2].y, verts[3].x, verts[3].y,
                                  verts[4].x, verts[4].y)
        elseif v.type == "t" then
            local verts = v.verts

            love.graphics.polygon("fill", verts[1].x, verts[1].y, verts[2].x,
                                  verts[2].y, verts[3].x, verts[3].y)
        elseif v.type == "o" then
            local verts = v.verts
            love.graphics.polygon("fill", verts[1].x, verts[1].y, verts[2].x,
                                  verts[2].y, verts[3].x, verts[3].y,
                                  verts[4].x, verts[4].y, verts[5].x,
                                  verts[5].y, verts[6].x, verts[6].y,
                                  verts[7].x, verts[7].y, verts[8].x, verts[8].y)

        end

    end

end
